"""
Build SSW scripts from Jinja 2 templates
"""
import os
import datetime
import pathlib
import subprocess
import tempfile

import jinja2
from scipy.io import readsav
from .filters import string_list_filter

from .read_config import defaults
from .util import SSWIDLError, IDLLicenseError
from .filters import *


class Environment(object):
    """
    Environment for running SSW and IDL scripts

    Parameters
    ----------
    ssw_packages : `list`, optional
        List of SSW packages to load, e.g. 'sdo/aia', 'chianti'
    ssw_paths : `list`, optional
        List of SSW paths to pass to `ssw_path`
    extra_paths : `list`, optional
        Additional paths to add to the IDL namespace. Note that these
        are appended to the front of the path such that they take
        precedence over the existing path.
    ssw_home : `str`, optional
        Root of SSW tree
    idl_home : `str`, optional
        Path to IDL executable
    filters : `dict`, optional
        Filters to use in scripts. This should be a dictionary where the key
        is the name of the filter and the value is the corresponding
        function.
    idl_only : `bool`, optional
        If True, do not do any setup associated with SSW. This is useful
        if your script has no SSW dependence.
    header_script : `str` or path-like, optional
        Script to run before script passed to ``run`` method. Can use any
        of the variables passed to ``run``.
    footer_script : `str` or path-like, optional
        Script to run after script passed to ``run`` method. Can use any
        of the variables passed to ``run``.
    """

    def __init__(self, ssw_packages=None, ssw_paths=None, extra_paths=None,
                 ssw_home=None, idl_home=None, filters=None, idl_only=False,
                 header=None, footer=None):
        self.ssw_packages = ssw_packages if ssw_packages is not None else []
        self.ssw_paths = ssw_paths if ssw_paths is not None else []
        self.extra_paths = extra_paths if extra_paths is not None else []
        self.env = jinja2.Environment(loader=jinja2.PackageLoader('hissw', 'templates'))
        self.env.filters['to_unit'] = units_filter
        self.env.filters['log10'] = log10_filter
        self.env.filters['string_list'] = string_list_filter
        self.env.filters['force_double_precision'] = force_double_precision_filter
        if filters is not None:
            for k, v in filters.items():
                self.env.filters[k] = v
        self.header = '' if header is None else header
        self.footer = '' if footer is None else footer
        self._setup_home(ssw_home, idl_home, idl_only=idl_only)

    def _setup_home(self, ssw_home, idl_home, idl_only=False):
        """
        Setup SSW and IDL home locations
        """
        self.ssw_home = defaults.get('ssw_home') if ssw_home is None else ssw_home
        if idl_only:
            self.ssw_home = None
        else:
            if self.ssw_home is None:
                raise ValueError('ssw_home must be set at instantiation or in the hisswrc file.')
        self.idl_home = defaults.get('idl_home') if idl_home is None else idl_home
        if self.idl_home is None:
            raise ValueError('idl_home must be set at instantiation or in the hisswrc file.')

    @property
    def executable(self):
        """
        Path to executable for running code
        """
        if self.ssw_home:
            return 'sswidl'
        else:
            return os.path.join(self.idl_home, 'bin', 'idl')

    def render_script(self, script, args):
        """
        Render custom IDL scripts from templates and input arguments
        """
        if isinstance(script, (str, pathlib.Path)) and os.path.isfile(script):
            with open(script, 'r') as f:
                script = f.read()
        if not isinstance(script, str):
            raise ValueError('Input script must either be a string or path to a script.')
        return self.env.from_string(script).render(**args)

    def custom_script(self, script, args):
        """
        Generate the script that will be executed
        """
        body = self.render_script(script, args)
        header = self.render_script(self.header, args)
        footer = self.render_script(self.footer, args)
        idl_script = f'{header}\n{body}\n{footer}'
        return idl_script

    def procedure_script(self, script, save_vars, save_filename):
        """
        Render inner procedure file
        """
        if save_vars is None:
            save_vars = []
        params = {'_script': script,
                  '_save_vars': save_vars,
                  '_save_filename': save_filename}
        return self.env.get_template('procedure.pro').render(**params)

    def command_script(self, procedure_filename):
        """
        Generate parent IDL script
        """
        params = {'ssw_paths': self.ssw_paths,
                  'extra_paths': self.extra_paths,
                  'procedure_filename': procedure_filename}
        return self.env.get_template('parent.pro').render(**params)

    def shell_script(self, command_filename):
        """
        Generate shell script for starting up SSWIDL
        """
        params = {'executable': self.executable,
                  'ssw_home': self.ssw_home,
                  'ssw_packages': self.ssw_packages,
                  'idl_home': self.idl_home,
                  'command_filename': command_filename}
        return self.env.get_template('startup.sh').render(**params)

    def run(self, script, args=None, save_vars=None, verbose=True, **kwargs):
        """
        Set up the SSWIDL environment and run the supplied scripts.

        Parameters
        ----------
        script : str
            Literal script or path to script file
        args : dict, optional
            Input arguments to script
        save_vars : list, optional
            Variables to save and return from the IDL namespace
        verbose : bool, optional
            If True, print STDERR and SDOUT. Otherwise it will be
            suppressed. This is useful for debugging.
        """
        args = {} if args is None else args
        # Expose the ssw_home variable in all scripts by default
        args.update({'ssw_home': self.ssw_home})
        with tempfile.TemporaryDirectory() as tmpdir:
            # Get filenames
            fn_template = os.path.join(
                tmpdir, '{name}_'+datetime.datetime.now().strftime('%Y%m%d-%H%M%S')+'.{ext}')
            save_filename = fn_template.format(name='idl_vars', ext='sav')
            procedure_filename = fn_template.format(name='idl_procedure', ext='pro')
            command_filename = fn_template.format(name='idl_script', ext='pro')
            shell_filename = fn_template.format(name='ssw_shell', ext='sh')
            # Render and save scripts
            idl_script = self.custom_script(script, args)
            with open(procedure_filename, 'w') as f:
                f.write(self.procedure_script(idl_script, save_vars, save_filename))
            with open(command_filename, 'w') as f:
                f.write(self.command_script(procedure_filename))
            with open(shell_filename, 'w') as f:
                f.write(self.shell_script(command_filename,))
            # Execute
            subprocess.call(['chmod', 'u+x', shell_filename])
            cmd_output = subprocess.run([shell_filename], shell=True, stderr=subprocess.PIPE,
                                        stdout=subprocess.PIPE)
            self._check_for_errors(cmd_output, verbose, **kwargs)
            results = readsav(save_filename)

        return results

    def _check_for_errors(self, output, verbose, **kwargs):
        """
        Check IDL output to try and decide if an error has occurred
        """
        stdout = output.stdout.decode('utf-8')
        stderr = output.stderr.decode('utf-8')
        # NOTE: For some reason, not only errors are output to stderr so we
        # have to check it for certain keywords to see if an error occurred
        if kwargs.get('raise_exceptions', True):
            if 'execution halted' in stderr.lower():
                raise SSWIDLError(stderr)
            if 'failed to acquire license' in stderr.lower():
                raise IDLLicenseError(stderr)
        if verbose:
            print(f'{stderr}\n{stdout}')



"""
Define some useful Jinja2 filters
"""
import numpy as np

__all__ = ['units_filter',
           'log10_filter',
           'string_list_filter',
           'force_double_precision_filter',]


def units_filter(quantity, unit):
    """
    Convert quantity to given units and extract value.
    """
    import astropy.units as u  # Avoid astropy as a hard dependency
    if not isinstance(quantity, u.Quantity):
        raise u.UnitsError(f'Value must be a quantity with units compatible with {unit}')
    return quantity.to_value(unit)


def log10_filter(value):
    """
    Take the base 10 log of a given value.
    """
    return np.log10(value)


def string_list_filter(string_list):
    """
    Double quote a list of strings.
    
    This is needed when passing in a list of strings to IDL as each string
    in the list will not be quoted when passed into the template.
    """
    return [f"'{s}'" for s in string_list]


def force_double_precision_filter(value):
    """
    Force a number (or array of numbers) to have double precision in IDL.

    hissw relies on string representations of Python objects when inserting
    values into the resulting IDL script. However, the default string representation
    of Python floats is truncated well below the actual floating point precision.
    See https://docs.python.org/3/tutorial/floatingpoint.html#floating-point-arithmetic-issues-and-limitations.
    Thus, to preserve precision, this filter uses the `as_integer_ratio` method
    to represent floating point values as division operations between integer values
    to ensure that precision is preserved when passing floating point values from
    IDL into Python. When used in combination with other filters, this filter should
    be used last as it forces a conversion to a list of strings.
    """
    if isinstance(value, (np.ndarray, list)):
        str_list = [force_double_precision_filter(x) for x in value]
        # NOTE: this has to be done manually because each entry is formatted as a
        # string such that the division is not evaluated in Python. However, we
        # want this to be inserted as an array of integer divison operations.
        return f"[{','.join(str_list)}]"
    else:
        # If it is neither an array or list, 
        a, b = value.as_integer_ratio()
        return f'{a}d / {b}d'


"""
Module level tests
"""
import pytest
import hissw
import numpy as np
import astropy.units as u
from hissw.util import SSWIDLError

run_kwargs = {'verbose': True}


@pytest.fixture
def idl_env(idl_home):
    return hissw.Environment(idl_home=idl_home, idl_only=True)


@pytest.fixture
def ssw_env(idl_home, ssw_home):
    return hissw.Environment(idl_home=idl_home,
                             ssw_home=ssw_home,
                             ssw_packages=['sdo/aia'],
                             ssw_paths=['aia'])


def test_exception(idl_env):
    """
    Test exception catching
    """
    with pytest.raises(SSWIDLError):
        _ = idl_env.run('foobar', **run_kwargs)


def test_no_args(idl_env):
    """
    No input arguments and no calls to SSW functions
    """
    script = '''
    n = 5
    i = REBIN(LINDGEN(n), n, n)
    j = REBIN(TRANSPOSE(LINDGEN(n)), n, n)
    array = (i GE j)
    '''
    results = idl_env.run(script, **run_kwargs)
    assert results['array'].shape == (5, 5)


def test_with_args(idl_env):
    """
    No calls to SSW functions
    """
    script = '''
    n = {{ n }}
    i = REBIN(LINDGEN(n), n, n)           
    j = REBIN(TRANSPOSE(LINDGEN(n)), n, n)
    array = (i GE j)
    '''
    n = 100
    results = idl_env.run(script, args={'n': n}, **run_kwargs)
    assert results['array'].shape == (n, n)


def test_aia_response_functions(ssw_env):
    """
    Compute AIA response functions using AIA packages in SSW
    """
    script = '''response = aia_get_response(/{{ flags | join(',/') }})
    ; Pull needed elements out of structure
    logt = response.logte
    resp94 = response.a94.tresp
    resp131 = response.a131.tresp
    resp171 = response.a171.tresp
    resp193 = response.a193.tresp
    resp211 = response.a211.tresp
    resp335 = response.a335.tresp
    '''
    args = {'flags': ['temp', 'dn', 'timedepend_date', 'evenorm']}
    results = ssw_env.run(script, args=args, **run_kwargs)
    for c  in [94, 131, 171, 193, 211, 335]:
        assert f'resp{c}' in results
        assert results[f'resp{c}'].shape == results['logt'].shape


def test_units_filter(idl_env):
    script = """
    foo = {{ foo | to_unit('K') }}
    """
    res = idl_env.run(script, args={'foo': 1*u.MK})
    assert np.allclose(res['foo'], 1e6)


def test_units_filter_exception(idl_env):
    script = """
    foo = {{ foo | to_unit('K') }}
    """
    with pytest.raises(u.UnitsError,
                       match='Value must be a quantity with units compatible with K'):
        _= idl_env.run(script, args={'foo': 1})


def test_units_filter_array(idl_env):
    script = """
    foo = {{ foo | to_unit('K') | list }}
    """
    res = idl_env.run(script, args={'foo': [1, 2]*u.MK})
    assert np.allclose(res['foo'], [1e6, 2e6])


def test_log10_filter(idl_env):
    script = """
    foo = {{ foo | log10 }}
    """
    foo = 1e6
    res = idl_env.run(script, args={'foo': foo})
    assert res['foo'] == np.log10(foo)


def test_string_list_filter(idl_env):
    script = """
    foo = {{ foo | string_list }}
    """
    foo = ['my', 'list', 'of', 'strings']
    res = idl_env.run(script, args={'foo': foo})
    assert [v.decode('utf-8') for v in res['foo']] == [f"'{f}'" for f in foo]


def test_default_ssw_var(ssw_env):
    script = """
    foo = '{{ ssw_home }}'
    """
    res = ssw_env.run(script)
    assert res['foo'].decode('utf-8') == ssw_env.ssw_home


def test_script_from_file(idl_env, tmp_path):
    script = '''foo = {{a}} + {{b}}'''
    script_file = tmp_path / 'test_script.pro'
    with open(script_file, 'w') as f:
        f.write(script)
    a = 1
    b = 2
    result = idl_env.run(script_file, args={'a':a, 'b': b})
    assert result['foo'] == (a + b)


def test_custom_filters(idl_home):
    filters = {
        'my_filter': lambda x: 'foo' if x < 0.5 else 'bar'
    }
    env = hissw.Environment(idl_home=idl_home, idl_only=True, filters=filters)
    script = '''
    a = '{{ a | my_filter }}'
    b = '{{ b | my_filter }}'
    '''
    args = {'a': 0.1, 'b': 0.6}
    result = env.run(script, args=args)
    assert result['a'].decode('utf-8') == filters['my_filter'](args['a'])
    assert result['b'].decode('utf-8') == filters['my_filter'](args['b'])


def test_invalid_script(idl_env):
    with pytest.raises(ValueError, match='Input script must either be a string or path to a script.'):
        _ = idl_env.run(None)


def test_custom_header_footer(idl_home):
    header = 'foo = {{ a }}'
    footer = 'bar = {{ a }} + {{ b }}'
    env_custom = hissw.Environment(idl_home=idl_home, idl_only=True,
                                   header=header, footer=footer)
    script = '''
    print, {{ a }}
    print, {{ b }}
    '''
    args = {'a': 1, 'b': 2}
    result = env_custom.run(script, args=args)
    assert result['foo'] == args['a']
    assert result['bar'] == args['a'] + args['b']


@pytest.mark.parametrize('var', [
    1 / 3,
    float(10),
    [1/3, 1/7, 0.1],
    np.random.rand(100),
    list(np.random.rand(10).astype(np.longdouble))
])
def test_force_double_precision_filter(var, idl_env):
    # This test ensures that floating point precision is conserved when passing
    # things from Python to IDL. See https://github.com/wtbarnes/hissw/issues/31
    result = idl_env.run(
            '''
            var = {{ var | force_double_precision }}
            var_size = size(var)
            ''',
            args={'var': var})
    assert u.allclose(var, result['var'], atol=0.0, rtol=0.0)
    # The result of IDL size has a variable number of entries depending on the
    # dimensionality of the input array
    # NOTE: 5 corresponds to double-precision floating point. See
    # https://www.l3harrisgeospatial.com/docs/make_array.html#TYPE
    assert result['var_size'][len(result['var'].shape) + 1] == 5


@pytest.mark.parametrize('var', [
    1 / 3 * u.s,
    float(10) * u.s,
    [1/3, 1/7, 0.1] * u.s,
    np.random.rand(100) * u.minute,
])
def test_force_double_precision_filter_with_quantity(var, idl_env):
    # This test ensures that floating point precision is conserved when passing
    # things from Python to IDL. See https://github.com/wtbarnes/hissw/issues/31
    result = idl_env.run('var = {{ var | to_unit("h") | force_double_precision }}',
                         args={'var': var})
    assert u.allclose(var.to_value('h'), result['var'], atol=0.0, rtol=0.0)

